@@ -1,13 +1,17 @@
-/* dnsmasq is Copyright (c) 2000-2006 Simon Kelley
+/* dnsmasq is Copyright (c) 2000-2007 Simon Kelley
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 dated June, 1991.
-
+   the Free Software Foundation; version 2 dated June, 1991, or
+   (at your option) version 3 dated 29 June, 2007.
+ 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
+     
+  You should have received a copy of the GNU General Public License
+  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "dnsmasq.h"
@@ -24,67 +28,81 @@
    main process.
 */
 
+#ifndef NO_FORK
+
+static void my_setenv(const char *name, const char *value, int *error);
+ 
 struct script_data
 {
   unsigned char action, hwaddr_len, hwaddr_type;
   unsigned char clid_len, hostname_len, uclass_len, vclass_len;
   struct in_addr addr;
+  unsigned int remaining_time;
 #ifdef HAVE_BROKEN_RTC
   unsigned int length;
 #else
   time_t expires;
 #endif
   unsigned char hwaddr[DHCP_CHADDR_MAX];
+  char interface[IF_NAMESIZE];
 };
 
-static struct script_data *buf;
-static size_t bytes_in_buf, buf_size;
+static struct script_data *buf = NULL;
+static size_t bytes_in_buf = 0, buf_size = 0;
 
-int create_helper(struct daemon *daemon, int log_fd)
+int create_helper(int event_fd, long max_fd)
 {
   pid_t pid;
   int i, pipefd[2];
   struct sigaction sigact;
-
-  buf = NULL;
-  buf_size = bytes_in_buf = 0;
+  struct passwd *ent_pw = NULL;
 
   if (!daemon->dhcp || !daemon->lease_change_command)
     return -1;
+  
+  if (daemon->scriptuser && !(ent_pw = getpwnam(daemon->scriptuser)))
+    {
+      /* too late to die() */
+      send_event(event_fd, EVENT_USER_ERR, 0);
+      return -1;
+    }
 
   /* create the pipe through which the main program sends us commands,
-   then fork our process. */
+     then fork our process. By now it's too late to die(), we just log 
+     any failure via the main process. */
   if (pipe(pipefd) == -1 || !fix_fd(pipefd[1]) || (pid = fork()) == -1)
-    return -1;
-  
+    {
+      send_event(event_fd, EVENT_PIPE_ERR, errno);
+      return -1;
+    }
+
   if (pid != 0)
     {
       close(pipefd[0]); /* close reader side */
       return pipefd[1];
     }
 
-  /* ignore SIGTERM, so that we can clean up when the main process gets hit */
+  /* ignore SIGTERM, so that we can clean up when the main process gets hit
+     and SIGALRM so that we can use sleep() */
   sigact.sa_handler = SIG_IGN;
   sigact.sa_flags = 0;
   sigemptyset(&sigact.sa_mask);
   sigaction(SIGTERM, &sigact, NULL);
+  sigaction(SIGALRM, &sigact, NULL);
 
   /* close all the sockets etc, we don't need them here */
-  for (i = 0; i < 64; i++)
-    if (i != STDOUT_FILENO && i != STDERR_FILENO && 
-	i != STDIN_FILENO && i != pipefd[0] && i != log_fd)
-      close(i);
+  for (max_fd--; max_fd > 0; max_fd--)
+    if (max_fd != STDOUT_FILENO && max_fd != STDERR_FILENO && 
+	max_fd != STDIN_FILENO && max_fd != pipefd[0] && max_fd != event_fd)
+      close(max_fd);
   
-  /* don't give our end of the pipe to our children */
-  if ((i = fcntl(pipefd[0], F_GETFD)) != -1)
-    fcntl(pipefd[0], F_SETFD, i | FD_CLOEXEC); 
-      
   /* loop here */
   while(1)
     {
       struct script_data data;
       char *p, *action_str, *hostname = NULL;
       unsigned char *buf = (unsigned char *)daemon->namebuff;
+      int err = 0;
 
       /* we read zero bytes when pipe closed: this is our signal to exit */ 
       if (!read_write(pipefd[0], (unsigned char *)&data, sizeof(data), 1))
@@ -130,32 +148,63 @@ int create_helper(struct daemon *daemon, int log_fd)
       if (!read_write(pipefd[0], buf, data.hostname_len + data.uclass_len + data.vclass_len, 1))
 	continue;
       
-      if ((pid = fork()) == -1)
-	continue;
+      /* possible fork errors are all temporary resource problems */
+      while ((pid = fork()) == -1 && (errno == EAGAIN || errno == ENOMEM))
+	sleep(2);
       
+      if (pid == -1)
+	continue;
+	  
       /* wait for child to complete */
       if (pid != 0)
 	{
-	  int status;
-	  waitpid(pid, &status, 0);
-	  if (WIFSIGNALED(status))
-	    my_syslog(LOG_WARNING, _("child process killed by signal %d"), WTERMSIG(status));
-	  else if (WIFEXITED(status) && WEXITSTATUS(status) != 0)
-	    my_syslog(LOG_WARNING, _("child process exited with status %d"), WEXITSTATUS(status));
+	  /* reap our children's children, if necessary */
+	  while (1)
+	    {
+	      int status;
+	      pid_t rc = wait(&status);
+	      
+	      if (rc == pid)
+		{
+		  /* On error send event back to main process for logging */
+		  if (WIFSIGNALED(status))
+		    send_event(event_fd, EVENT_KILLED, WTERMSIG(status));
+		  else if (WIFEXITED(status) && WEXITSTATUS(status) != 0)
+		    send_event(event_fd, EVENT_EXITED, WEXITSTATUS(status));
+		  break;
+		}
+	      
+	      if (rc == -1 && errno != EINTR)
+		break;
+	    }
+	  
 	  continue;
 	}
       
+      /* Change ownership of child, if this fails, log and ditch, 
+	 rather than running wrong user */
+      if (ent_pw)
+	{
+	  gid_t dummy;
+	  if (setgroups(0, &dummy) == -1 || 
+	      setgid(ent_pw->pw_gid) == -1 || 
+	      setuid(ent_pw->pw_uid) == -1)
+	    {
+	      send_event(event_fd, EVENT_USER_ERR, errno);
+	      _exit(0);
+	    }
+	}
+
       if (data.clid_len != 0)
-	setenv("DNSMASQ_CLIENT_ID", daemon->packet, 1);
-      else
-	unsetenv("DNSMASQ_CLIENT_ID");
-      
+	my_setenv("DNSMASQ_CLIENT_ID", daemon->packet, &err);
+
+      if (strlen(data.interface) != 0)
+	my_setenv("DNSMASQ_INTERFACE", data.interface, &err);
+            
 #ifdef HAVE_BROKEN_RTC
-      setenv("DNSMASQ_LEASE_LENGTH", daemon->dhcp_buff2, 1);
-      unsetenv("DNSMASQ_LEASE_EXPIRES");
+      my_setenv("DNSMASQ_LEASE_LENGTH", daemon->dhcp_buff2, &err);
 #else
-      setenv("DNSMASQ_LEASE_EXPIRES", daemon->dhcp_buff2, 1); 
-      unsetenv("DNSMASQ_LEASE_LENGTH");
+      my_setenv("DNSMASQ_LEASE_EXPIRES", daemon->dhcp_buff2, &err); 
 #endif
       
       if (data.vclass_len != 0)
@@ -164,11 +213,9 @@ int create_helper(struct daemon *daemon, int log_fd)
 	  /* cannot have = chars in env - truncate if found . */
 	  if ((p = strchr((char *)buf, '=')))
 	    *p = 0;
-	  setenv("DNSMASQ_VENDOR_CLASS", (char *)buf, 1);
+	  my_setenv("DNSMASQ_VENDOR_CLASS", (char *)buf, &err);
 	  buf += data.vclass_len;
 	}
-      else 
-	unsetenv("DNSMASQ_VENDOR_CLASS");
       
       if (data.uclass_len != 0)
 	{
@@ -183,45 +230,80 @@ int create_helper(struct daemon *daemon, int log_fd)
 	      if (strlen((char *)buf) != 0)
 		{
 		  sprintf(daemon->dhcp_buff2, "DNSMASQ_USER_CLASS%i", i++);
-		  setenv(daemon->dhcp_buff2, (char *)buf, 1);
+		  my_setenv(daemon->dhcp_buff2, (char *)buf, &err);
 		}
 	      buf += len;
 	    }
 	}
       
+      sprintf(daemon->dhcp_buff2, "%u ", data.remaining_time);
+      my_setenv("DNSMASQ_TIME_REMAINING", daemon->dhcp_buff2, &err);
+      
       if (data.hostname_len != 0)
 	{
 	  hostname = (char *)buf;
 	  hostname[data.hostname_len - 1] = 0;
-	  canonicalise(hostname);
+	  if (!canonicalise(hostname))
+	    hostname = NULL;
 	}
       
       if (data.action == ACTION_OLD_HOSTNAME && hostname)
 	{
-	  setenv("DNSMASQ_OLD_HOSTNAME", hostname, 1);
+	  my_setenv("DNSMASQ_OLD_HOSTNAME", hostname, &err);
 	  hostname = NULL;
 	}
-      else
-	unsetenv("DNSMASQ_OLD_HOSTNAME");
-      
+
+      /* we need to have the event_fd around if exec fails */
+      if ((i = fcntl(event_fd, F_GETFD)) != -1)
+	fcntl(event_fd, F_SETFD, i | FD_CLOEXEC);
+      close(pipefd[0]);
+
       p =  strrchr(daemon->lease_change_command, '/');
-      execl(daemon->lease_change_command, 
-	    p ? p+1 : daemon->lease_change_command,
-	    action_str, daemon->dhcp_buff, inet_ntoa(data.addr), hostname, (char*)NULL);
-      
-      /* log socket should still be open, right? */
-      my_syslog(LOG_ERR, _("failed to execute %s: %s"), 
-		daemon->lease_change_command, strerror(errno));
+      if (err == 0)
+	{
+	  execl(daemon->lease_change_command, 
+		p ? p+1 : daemon->lease_change_command,
+		action_str, daemon->dhcp_buff, inet_ntoa(data.addr), hostname, (char*)NULL);
+	  err = errno;
+	}
+      /* failed, send event so the main process logs the problem */
+      send_event(event_fd, EVENT_EXEC_ERR, err);
       _exit(0); 
     }
 }
 
+static void my_setenv(const char *name, const char *value, int *error)
+{
+  if (*error == 0)
+    {
+#if defined(HAVE_SOLARIS_NETWORK) && !defined(HAVE_SOLARIS_PRIVS)
+      /* old Solaris is missing setenv..... */
+      char *p;
+      
+      if (!(p = malloc(strlen(name) + strlen(value) + 2)))
+	*error = ENOMEM;
+      else
+	{
+	  strcpy(p, name);
+	  strcat(p, "=");
+	  strcat(p, value);
+	  
+	  if (putenv(p) != 0)
+	    *error = errno;
+	}
+#else
+      if (setenv(name, value, 1) != 0)
+	*error = errno;
+#endif
+    }
+}
+ 
 /* pack up lease data into a buffer */    
-void queue_script(struct daemon *daemon, int action, struct dhcp_lease *lease, char *hostname)
+void queue_script(int action, struct dhcp_lease *lease, char *hostname, time_t now)
 {
   unsigned char *p;
   size_t size;
-  unsigned int hostname_len = 0, clid_len = 0, vclass_len = 0, uclass_len = 0;
+  unsigned int i, hostname_len = 0, clid_len = 0, vclass_len = 0, uclass_len = 0;
 
   /* no script */
   if (daemon->helperfd == -1)
@@ -246,7 +328,7 @@ void queue_script(struct daemon *daemon, int action, struct dhcp_lease *lease, c
       if (size < sizeof(struct script_data) + 200)
 	size = sizeof(struct script_data) + 200;
 
-      if (!(new = malloc(size)))
+      if (!(new = whine_malloc(size)))
 	return;
       if (buf)
 	free(buf);
@@ -263,34 +345,50 @@ void queue_script(struct daemon *daemon, int action, struct dhcp_lease *lease, c
   buf->hostname_len = hostname_len;
   buf->addr = lease->addr;
   memcpy(buf->hwaddr, lease->hwaddr, lease->hwaddr_len);
+  buf->interface[0] = 0;
+#ifdef HAVE_LINUX_NETWORK
+  if (lease->last_interface != 0)
+    {
+      struct ifreq ifr;
+      ifr.ifr_ifindex = lease->last_interface;
+      if (ioctl(daemon->dhcpfd, SIOCGIFNAME, &ifr) != -1)
+	strncpy(buf->interface, ifr.ifr_name, IF_NAMESIZE);
+    }
+#else
+  if (lease->last_interface != 0)
+    if_indextoname(lease->last_interface, buf->interface);
+#endif
+  
 #ifdef HAVE_BROKEN_RTC 
   buf->length = lease->length;
 #else
   buf->expires = lease->expires;
 #endif
- 
+  buf->remaining_time = (unsigned int)difftime(lease->expires, now);
+
   p = (unsigned char *)(buf+1);
-  if (buf->clid_len != 0)
+  if (clid_len != 0)
     {
       memcpy(p, lease->clid, clid_len);
       p += clid_len;
     }
-  if (buf->vclass_len != 0)
+  if (vclass_len != 0)
     {
       memcpy(p, lease->vendorclass, vclass_len);
       p += vclass_len;
     }
-  if (buf->uclass_len != 0)
+  if (uclass_len != 0)
     {
       memcpy(p, lease->userclass, uclass_len);
       p += uclass_len;
     }
-  if (buf->hostname_len != 0)
-    {
-      memcpy(p, hostname, hostname_len);
-      p += hostname_len;
-    }
-
+  /* substitute * for space */
+  for (i = 0; i < hostname_len; i++)
+    if ((daemon->options & OPT_LEASE_RO) && hostname[i] == ' ')
+      *(p++) = '*';
+    else
+      *(p++) = hostname[i];
+  
   bytes_in_buf = p - (unsigned char *)buf;
 }
 
@@ -299,7 +397,7 @@ int helper_buf_empty(void)
   return bytes_in_buf == 0;
 }
 
-void helper_write(struct daemon *daemon)
+void helper_write(void)
 {
   ssize_t rc;
 
@@ -320,5 +418,6 @@ void helper_write(struct daemon *daemon)
     }
 }
 
+#endif
 
 
