@@ -1,18 +1,22 @@
-/* dnsmasq is Copyright (c) 2000 - 2006 Simon Kelley
+/* dnsmasq is Copyright (c) 2000-2007 Simon Kelley
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 dated June, 1991.
-
+   the Free Software Foundation; version 2 dated June, 1991, or
+   (at your option) version 3 dated 29 June, 2007.
+ 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
+     
+  You should have received a copy of the GNU General Public License
+  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "dnsmasq.h"
 
-int iface_check(struct daemon *daemon, int family, struct all_addr *addr, 
+int iface_check(int family, struct all_addr *addr, 
 		struct ifreq *ifr, int *indexp)
 {
   struct iname *tmp;
@@ -23,8 +27,8 @@ int iface_check(struct daemon *daemon, int family, struct all_addr *addr,
 
   if (indexp)
     {
-#if defined(__FreeBSD__) || defined(__DragonFly__)
-      /* One form of bridging on FreeBSD has the property that packets
+#ifdef HAVE_BSD_BRIDGE
+      /* One form of bridging on BSD has the property that packets
 	 can be recieved on bridge interfaces which do not have an IP address.
 	 We allow these to be treated as aliases of another interface which does have
 	 an IP address with --dhcp-bridge=interface,alias,alias */
@@ -84,7 +88,7 @@ int iface_check(struct daemon *daemon, int family, struct all_addr *addr,
   return ret; 
 }
       
-static int iface_allowed(struct daemon *daemon, struct irec **irecp, int if_index, 
+static int iface_allowed(struct irec **irecp, int if_index, 
 			 union mysockaddr *addr, struct in_addr netmask) 
 {
   struct irec *iface;
@@ -134,9 +138,10 @@ static int iface_allowed(struct daemon *daemon, struct irec **irecp, int if_inde
 	    break;
 	  }
       
-      if (!lo && (lo = malloc(sizeof(struct iname))))
+      if (!lo && 
+	  (lo = whine_malloc(sizeof(struct iname))) &&
+	  (lo->name = whine_malloc(strlen(ifr.ifr_name)+1)))
 	{
-	  lo->name = safe_malloc(strlen(ifr.ifr_name)+1);
 	  strcpy(lo->name, ifr.ifr_name);
 	  lo->isloop = lo->used = 1;
 	  lo->next = daemon->if_names;
@@ -145,7 +150,7 @@ static int iface_allowed(struct daemon *daemon, struct irec **irecp, int if_inde
     }
   
   if (addr->sa.sa_family == AF_INET &&
-      !iface_check(daemon, AF_INET, (struct all_addr *)&addr->in.sin_addr, &ifr, NULL))
+      !iface_check(AF_INET, (struct all_addr *)&addr->in.sin_addr, &ifr, NULL))
     return 1;
   
   for (tmp = daemon->dhcp_except; tmp; tmp = tmp->next)
@@ -154,12 +159,12 @@ static int iface_allowed(struct daemon *daemon, struct irec **irecp, int if_inde
   
 #ifdef HAVE_IPV6
   if (addr->sa.sa_family == AF_INET6 &&
-      !iface_check(daemon, AF_INET6, (struct all_addr *)&addr->in6.sin6_addr, &ifr, NULL))
+      !iface_check(AF_INET6, (struct all_addr *)&addr->in6.sin6_addr, &ifr, NULL))
     return 1;
 #endif
 
   /* add to list */
-  if ((iface = malloc(sizeof(struct irec))))
+  if ((iface = whine_malloc(sizeof(struct irec))))
     {
       iface->addr = *addr;
       iface->netmask = netmask;
@@ -174,7 +179,7 @@ static int iface_allowed(struct daemon *daemon, struct irec **irecp, int if_inde
 }
 
 #ifdef HAVE_IPV6
-static int iface_allowed_v6(struct daemon *daemon, struct in6_addr *local, 
+static int iface_allowed_v6(struct in6_addr *local, 
 			    int scope, int if_index, void *vparam)
 {
   union mysockaddr addr;
@@ -191,11 +196,11 @@ static int iface_allowed_v6(struct daemon *daemon, struct in6_addr *local,
   addr.in6.sin6_port = htons(daemon->port);
   addr.in6.sin6_scope_id = scope;
   
-  return iface_allowed(daemon, (struct irec **)vparam, if_index, &addr, netmask);
+  return iface_allowed((struct irec **)vparam, if_index, &addr, netmask);
 }
 #endif
 
-static int iface_allowed_v4(struct daemon *daemon, struct in_addr local, int if_index, 
+static int iface_allowed_v4(struct in_addr local, int if_index, 
 			    struct in_addr netmask, struct in_addr broadcast, void *vparam)
 {
   union mysockaddr addr;
@@ -209,30 +214,28 @@ static int iface_allowed_v4(struct daemon *daemon, struct in_addr local, int if_
   addr.in.sin_addr = local;
   addr.in.sin_port = htons(daemon->port);
 
-  return iface_allowed(daemon, (struct irec **)vparam, if_index, &addr, netmask);
+  return iface_allowed((struct irec **)vparam, if_index, &addr, netmask);
 }
    
 
-int enumerate_interfaces(struct daemon *daemon)
+int enumerate_interfaces(void)
 {
 #ifdef HAVE_IPV6
-  return iface_enumerate(daemon, &daemon->interfaces, iface_allowed_v4, iface_allowed_v6);
+  return iface_enumerate(&daemon->interfaces, iface_allowed_v4, iface_allowed_v6);
 #else
-  return iface_enumerate(daemon, &daemon->interfaces, iface_allowed_v4, NULL);
+  return iface_enumerate(&daemon->interfaces, iface_allowed_v4, NULL);
 #endif
 }
 
-/* set NONBLOCK and CLOEXEC bits on fd: See Stevens 16.6 */
+/* set NONBLOCK bit on fd: See Stevens 16.6 */
 int fix_fd(int fd)
 {
   int flags;
 
   if ((flags = fcntl(fd, F_GETFL)) == -1 ||
-      fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1 ||
-      (flags = fcntl(fd, F_GETFD)) == -1 ||
-      fcntl(fd, F_SETFD, flags | FD_CLOEXEC) == -1)
+      fcntl(fd, F_SETFL, flags | O_NONBLOCK) == -1)
     return 0;
-
+  
   return 1;
 }
 
@@ -289,52 +292,55 @@ static int create_ipv6_listener(struct listener **link, int port)
 }
 #endif
 
-struct listener *create_wildcard_listeners(int port, int have_tftp)
+struct listener *create_wildcard_listeners(void)
 {
   union mysockaddr addr;
   int opt = 1;
   struct listener *l, *l6 = NULL;
-  int tcpfd, fd, tftpfd = -1;
+  int tcpfd = -1, fd = -1, tftpfd = -1;
 
   memset(&addr, 0, sizeof(addr));
   addr.in.sin_family = AF_INET;
   addr.in.sin_addr.s_addr = INADDR_ANY;
-  addr.in.sin_port = htons(port);
+  addr.in.sin_port = htons(daemon->port);
 #ifdef HAVE_SOCKADDR_SA_LEN
   addr.in.sin_len = sizeof(struct sockaddr_in);
 #endif
 
-  if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1 ||
-      (tcpfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
-    return NULL;
-  
-  if (setsockopt(tcpfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
-      bind(tcpfd, (struct sockaddr *)&addr, sa_len(&addr)) == -1 ||
-      listen(tcpfd, 5) == -1 ||
-      !fix_fd(tcpfd) ||
+  if (daemon->port != 0)
+    {
+      
+      if ((fd = socket(AF_INET, SOCK_DGRAM, 0)) == -1 ||
+	  (tcpfd = socket(AF_INET, SOCK_STREAM, 0)) == -1)
+	return NULL;
+      
+      if (setsockopt(tcpfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
+	  bind(tcpfd, (struct sockaddr *)&addr, sa_len(&addr)) == -1 ||
+	  listen(tcpfd, 5) == -1 ||
+	  !fix_fd(tcpfd) ||
 #ifdef HAVE_IPV6
-      !create_ipv6_listener(&l6, port) ||
+	  !create_ipv6_listener(&l6, daemon->port) ||
 #endif
-      setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
-      !fix_fd(fd) ||
+	  setsockopt(fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
+	  !fix_fd(fd) ||
 #if defined(HAVE_LINUX_NETWORK) 
-      setsockopt(fd, SOL_IP, IP_PKTINFO, &opt, sizeof(opt)) == -1 ||
+	  setsockopt(fd, SOL_IP, IP_PKTINFO, &opt, sizeof(opt)) == -1 ||
 #elif defined(IP_RECVDSTADDR) && defined(IP_RECVIF)
-      setsockopt(fd, IPPROTO_IP, IP_RECVDSTADDR, &opt, sizeof(opt)) == -1 ||
-      setsockopt(fd, IPPROTO_IP, IP_RECVIF, &opt, sizeof(opt)) == -1 ||
+	  setsockopt(fd, IPPROTO_IP, IP_RECVDSTADDR, &opt, sizeof(opt)) == -1 ||
+	  setsockopt(fd, IPPROTO_IP, IP_RECVIF, &opt, sizeof(opt)) == -1 ||
 #endif 
-      bind(fd, (struct sockaddr *)&addr, sa_len(&addr)) == -1)
-    return NULL;
-
+	  bind(fd, (struct sockaddr *)&addr, sa_len(&addr)) == -1)
+	return NULL;
+    }
+  
 #ifdef HAVE_TFTP
-  if (have_tftp)
+  if (daemon->options & OPT_TFTP)
     {
       addr.in.sin_port = htons(TFTP_PORT);
       if ((tftpfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1)
 	return NULL;
       
-      if (setsockopt(tftpfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
-	  !fix_fd(tftpfd) ||
+      if (!fix_fd(tftpfd) ||
 #if defined(HAVE_LINUX_NETWORK) 
 	  setsockopt(tftpfd, SOL_IP, IP_PKTINFO, &opt, sizeof(opt)) == -1 ||
 #elif defined(IP_RECVDSTADDR) && defined(IP_RECVIF)
@@ -356,7 +362,7 @@ struct listener *create_wildcard_listeners(int port, int have_tftp)
   return l;
 }
 
-struct listener *create_bound_listeners(struct daemon *daemon)
+struct listener *create_bound_listeners(void)
 {
   struct listener *listeners = NULL;
   struct irec *iface;
@@ -369,49 +375,52 @@ struct listener *create_bound_listeners(struct daemon *daemon)
       new->iface = iface;
       new->next = listeners;
       new->tftpfd = -1;
+      new->tcpfd = -1;
+      new->fd = -1;
 
-      if ((new->tcpfd = socket(iface->addr.sa.sa_family, SOCK_STREAM, 0)) == -1 ||
-	  (new->fd = socket(iface->addr.sa.sa_family, SOCK_DGRAM, 0)) == -1 ||
-	  setsockopt(new->fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
-	  setsockopt(new->tcpfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
-	  !fix_fd(new->tcpfd) ||
-	  !fix_fd(new->fd))
-	die(_("failed to create listening socket: %s"), NULL);
-      
-#ifdef HAVE_IPV6
-      if (iface->addr.sa.sa_family == AF_INET6)
-	{
-	  if (setsockopt(new->fd, IPV6_LEVEL, IPV6_V6ONLY, &opt, sizeof(opt)) == -1 ||
-	      setsockopt(new->tcpfd, IPV6_LEVEL, IPV6_V6ONLY, &opt, sizeof(opt)) == -1)
-	    die(_("failed to set IPV6 options on listening socket: %s"), NULL);
-	}
-#endif
-      
-      if (bind(new->tcpfd, &iface->addr.sa, sa_len(&iface->addr)) == -1 ||
-	  bind(new->fd, &iface->addr.sa, sa_len(&iface->addr)) == -1)
+      if (daemon->port != 0)
 	{
+	  if ((new->tcpfd = socket(iface->addr.sa.sa_family, SOCK_STREAM, 0)) == -1 ||
+	      (new->fd = socket(iface->addr.sa.sa_family, SOCK_DGRAM, 0)) == -1 ||
+	      setsockopt(new->fd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
+	      setsockopt(new->tcpfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
+	      !fix_fd(new->tcpfd) ||
+	      !fix_fd(new->fd))
+	    die(_("failed to create listening socket: %s"), NULL, EC_BADNET);
+	  
 #ifdef HAVE_IPV6
-	  if (iface->addr.sa.sa_family == AF_INET6 && errno == ENODEV)
+	  if (iface->addr.sa.sa_family == AF_INET6)
 	    {
-	      close(new->tcpfd);
-	      close(new->fd);
-	      free(new);
+	      if (setsockopt(new->fd, IPV6_LEVEL, IPV6_V6ONLY, &opt, sizeof(opt)) == -1 ||
+		  setsockopt(new->tcpfd, IPV6_LEVEL, IPV6_V6ONLY, &opt, sizeof(opt)) == -1)
+		die(_("failed to set IPV6 options on listening socket: %s"), NULL, EC_BADNET);
 	    }
-	  else
 #endif
+	  
+	  if (bind(new->tcpfd, &iface->addr.sa, sa_len(&iface->addr)) == -1 ||
+	      bind(new->fd, &iface->addr.sa, sa_len(&iface->addr)) == -1)
 	    {
-	      prettyprint_addr(&iface->addr, daemon->namebuff);
-	      die(_("failed to bind listening socket for %s: %s"), 
-		  daemon->namebuff);
+#ifdef HAVE_IPV6
+	      if (iface->addr.sa.sa_family == AF_INET6 && (errno == ENODEV || errno == EADDRNOTAVAIL))
+		{
+		  close(new->tcpfd);
+		  close(new->fd);
+		  free(new);
+		  new = NULL;
+		}
+	      else
+#endif
+		{
+		  prettyprint_addr(&iface->addr, daemon->namebuff);
+		  die(_("failed to bind listening socket for %s: %s"), 
+		      daemon->namebuff, EC_BADNET);
+		}
 	    }
+	  else if (listen(new->tcpfd, 5) == -1)
+	    die(_("failed to listen on socket: %s"), NULL, EC_BADNET);
 	}
-      else
-	 {
-	   listeners = new;     
-	   if (listen(new->tcpfd, 5) == -1)
-	     die(_("failed to listen on socket: %s"), NULL);
-	 }
 
+#ifdef HAVE_TFTP
       if ((daemon->options & OPT_TFTP) && iface->addr.sa.sa_family == AF_INET && iface->dhcp_ok)
 	{
 	  short save = iface->addr.in.sin_port;
@@ -420,26 +429,59 @@ struct listener *create_bound_listeners(struct daemon *daemon)
 	      setsockopt(new->tftpfd, SOL_SOCKET, SO_REUSEADDR, &opt, sizeof(opt)) == -1 ||
 	      !fix_fd(new->tftpfd) ||
 	      bind(new->tftpfd, &iface->addr.sa, sa_len(&iface->addr)) == -1)
-	    die(_("failed to create TFTP socket: %s"), NULL);
+	    die(_("failed to create TFTP socket: %s"), NULL, EC_BADNET);
 	  iface->addr.in.sin_port = save;
 	}
+#endif
+
+      if (new)
+	listeners = new;
     }
 
   return listeners;
 }
 
-struct serverfd *allocate_sfd(union mysockaddr *addr, struct serverfd **sfds)
+int local_bind(int fd, union mysockaddr *addr, char *intname, int is_tcp)
+{
+  union mysockaddr addr_copy = *addr;
+
+  /* cannot set source _port_ for TCP connections. */
+  if (is_tcp)
+    {
+      if (addr_copy.sa.sa_family == AF_INET)
+	addr_copy.in.sin_port = 0;
+#ifdef HAVE_IPV6
+      else
+	addr_copy.in6.sin6_port = 0;
+#endif
+    }
+  
+  if (bind(fd, (struct sockaddr *)&addr_copy, sa_len(&addr_copy)) == -1)
+    return 0;
+    
+#if defined(SO_BINDTODEVICE)
+  if (strlen(intname) != 0 &&
+      setsockopt(fd, SOL_SOCKET, SO_BINDTODEVICE, intname, sizeof(intname)) == -1)
+    return 0;
+#endif
+
+  return 1;
+}
+
+static struct serverfd *allocate_sfd(union mysockaddr *addr, char *intname)
 {
   struct serverfd *sfd;
+  int errsave;
 
   /* may have a suitable one already */
-  for (sfd = *sfds; sfd; sfd = sfd->next )
-    if (sockaddr_isequal(&sfd->source_addr, addr))
+  for (sfd = daemon->sfds; sfd; sfd = sfd->next )
+    if (sockaddr_isequal(&sfd->source_addr, addr) &&
+	strcmp(intname, sfd->interface) == 0)
       return sfd;
   
   /* need to make a new one. */
   errno = ENOMEM; /* in case malloc fails. */
-  if (!(sfd = malloc(sizeof(struct serverfd))))
+  if (!(sfd = whine_malloc(sizeof(struct serverfd))))
     return NULL;
   
   if ((sfd->fd = socket(addr->sa.sa_family, SOCK_DGRAM, 0)) == -1)
@@ -448,24 +490,69 @@ struct serverfd *allocate_sfd(union mysockaddr *addr, struct serverfd **sfds)
       return NULL;
     }
   
-  if (bind(sfd->fd, (struct sockaddr *)addr, sa_len(addr)) == -1 ||
-      !fix_fd(sfd->fd))
-    {
-      int errsave = errno; /* save error from bind. */
+  if (!local_bind(sfd->fd, addr, intname, 0) || !fix_fd(sfd->fd))
+    { 
+      errsave = errno; /* save error from bind. */
       close(sfd->fd);
       free(sfd);
       errno = errsave;
       return NULL;
     }
-  
+    
+  strcpy(sfd->interface, intname); 
   sfd->source_addr = *addr;
-  sfd->next = *sfds;
-  *sfds = sfd;
-  
-  return sfd;
+  sfd->next = daemon->sfds;
+  daemon->sfds = sfd;
+  return sfd; 
 }
 
-void check_servers(struct daemon *daemon)
+/* create upstream sockets during startup, before root is dropped which may be needed
+   this allows query_port to be a low port and interface binding */
+void pre_allocate_sfds(void)
+{
+  struct server *srv;
+  
+  if (daemon->query_port != 0)
+    {
+      union  mysockaddr addr;
+      memset(&addr, 0, sizeof(addr));
+      addr.in.sin_family = AF_INET;
+      addr.in.sin_addr.s_addr = INADDR_ANY;
+      addr.in.sin_port = htons(daemon->query_port);
+#ifdef HAVE_SOCKADDR_SA_LEN
+      addr.in.sin_len = sizeof(struct sockaddr_in);
+#endif
+      allocate_sfd(&addr, "");
+#ifdef HAVE_IPV6
+      memset(&addr, 0, sizeof(addr));
+      addr.in6.sin6_family = AF_INET6;
+      addr.in6.sin6_addr = in6addr_any;
+      addr.in6.sin6_port = htons(daemon->query_port);
+#ifdef HAVE_SOCKADDR_SA_LEN
+      addr.in6.sin6_len = sizeof(struct sockaddr_in6);
+#endif
+      allocate_sfd(&addr, "");
+#endif
+    }
+  
+  for (srv = daemon->servers; srv; srv = srv->next)
+    if (!(srv->flags & (SERV_LITERAL_ADDRESS | SERV_NO_ADDR)) &&
+	!allocate_sfd(&srv->source_addr, srv->interface) &&
+	(daemon->options & OPT_NOWILD))
+      {
+	prettyprint_addr(&srv->addr, daemon->namebuff);
+	if (strlen(srv->interface) != 0)
+	  {
+	    strcat(daemon->namebuff, " ");
+	    strcat(daemon->namebuff, srv->interface);
+	  }
+	die(_("failed to bind server socket for %s: %s"),
+	    daemon->namebuff, EC_BADNET);
+      }  
+}
+
+
+void check_servers(void)
 {
   struct irec *iface;
   struct server *new, *tmp, *ret = NULL;
@@ -498,7 +585,7 @@ void check_servers(struct daemon *daemon)
 	    }
 	  
 	  /* Do we need a socket set? */
-	  if (!new->sfd && !(new->sfd = allocate_sfd(&new->source_addr, &daemon->sfds)))
+	  if (!new->sfd && !(new->sfd = allocate_sfd(&new->source_addr, new->interface)))
 	    {
 	      my_syslog(LOG_WARNING, 
 			_("ignoring nameserver %s - cannot make/bind socket: %s"),
@@ -515,16 +602,20 @@ void check_servers(struct daemon *daemon)
       if (new->flags & (SERV_HAS_DOMAIN | SERV_FOR_NODOTS))
 	{
 	  char *s1, *s2;
-	  if (new->flags & SERV_HAS_DOMAIN)
-	    s1 = _("domain"), s2 = new->domain;
+	  if (!(new->flags & SERV_HAS_DOMAIN))
+	    s1 = _("unqualified"), s2 = _("names");
+	  else if (strlen(new->domain) == 0)
+	    s1 = _("default"), s2 = "";
 	  else
-	    s1 = _("unqualified"), s2 = _("domains");
+	    s1 = _("domain"), s2 = new->domain;
 	  
 	  if (new->flags & SERV_NO_ADDR)
 	    my_syslog(LOG_INFO, _("using local addresses only for %s %s"), s1, s2);
 	  else if (!(new->flags & SERV_LITERAL_ADDRESS))
 	    my_syslog(LOG_INFO, _("using nameserver %s#%d for %s %s"), daemon->namebuff, port, s1, s2);
 	}
+      else if (strlen(new->interface) != 0)
+	my_syslog(LOG_INFO, _("using nameserver %s#%d(via %s)"), daemon->namebuff, port, new->interface); 
       else
 	my_syslog(LOG_INFO, _("using nameserver %s#%d"), daemon->namebuff, port); 
     }
@@ -534,7 +625,7 @@ void check_servers(struct daemon *daemon)
 
 /* Return zero if no servers found, in that case we keep polling.
    This is a protection against an update-time/write race on resolv.conf */
-int reload_servers(char *fname, struct daemon *daemon)
+int reload_servers(char *fname)
 {
   FILE *f;
   char *line;
@@ -561,7 +652,7 @@ int reload_servers(char *fname, struct daemon *daemon)
 	  serv->next = old_servers;
 	  old_servers = serv; 
 	  /* forward table rules reference servers, so have to blow them away */
-	  server_gone(daemon, serv);
+	  server_gone(serv);
 	}
       else
 	{
@@ -576,7 +667,9 @@ int reload_servers(char *fname, struct daemon *daemon)
       union mysockaddr addr, source_addr;
       char *token = strtok(line, " \t\n\r");
       
-      if (!token || strcmp(token, "nameserver") != 0)
+      if (!token)
+	continue;
+      if (strcmp(token, "nameserver") != 0 && strcmp(token, "server") != 0)
 	continue;
       if (!(token = strtok(NULL, " \t\n\r")))
 	continue;
@@ -614,7 +707,7 @@ int reload_servers(char *fname, struct daemon *daemon)
 	  serv = old_servers;
 	  old_servers = old_servers->next;
 	}
-      else if (!(serv = malloc(sizeof (struct server))))
+      else if (!(serv = whine_malloc(sizeof (struct server))))
 	continue;
       
       /* this list is reverse ordered: 
@@ -624,9 +717,10 @@ int reload_servers(char *fname, struct daemon *daemon)
       serv->addr = addr;
       serv->source_addr = source_addr;
       serv->domain = NULL;
+      serv->interface[0] = 0;
       serv->sfd = NULL;
       serv->flags = SERV_FROM_RESOLV;
-
+      serv->queries = serv->failed_queries = 0;
       gotone = 1;
     }
   
@@ -646,7 +740,7 @@ int reload_servers(char *fname, struct daemon *daemon)
 
 
 /* Use an IPv4 listener socket for ioctling */
-struct in_addr get_ifaddr(struct daemon* daemon, char *intr)
+struct in_addr get_ifaddr(char *intr)
 {
   struct listener *l;
   struct ifreq ifr;
