@@ -2,12 +2,16 @@
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
-   the Free Software Foundation; version 2 dated June, 1991.
-
+   the Free Software Foundation; version 2 dated June, 1991, or
+   (at your option) version 3 dated 29 June, 2007.
+ 
    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
+     
+  You should have received a copy of the GNU General Public License
+  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
 
 #include "dnsmasq.h"
@@ -67,24 +71,23 @@
 #define option_ptr(opt) ((void *)&(((unsigned char *)(opt))[2]))
 
 static int sanitise(unsigned char *opt, char *buf);
-static unsigned int calc_time(struct dhcp_context *context, struct dhcp_config *config, 
-			      struct dhcp_lease *lease, unsigned char *opt, time_t now);
+static unsigned int calc_time(struct dhcp_context *context, struct dhcp_config *config, unsigned char *opt);
 static void option_put(struct dhcp_packet *mess, unsigned char *end, int opt, int len, unsigned int val);
 static void option_put_string(struct dhcp_packet *mess, unsigned char *end, 
 			      int opt, char *string, int null_term);
 static struct in_addr option_addr(unsigned char *opt);
 static unsigned int option_uint(unsigned char *opt, int size);
-static void log_packet(struct daemon *daemon, char *type, void *addr, 
+static void log_packet(char *type, void *addr, 
 		       unsigned char *ext_mac, int mac_len, char *interface, char *string);
 static unsigned char *option_find(struct dhcp_packet *mess, size_t size, int opt_type, int minsize);
 static unsigned char *option_find1(unsigned char *p, unsigned char *end, int opt, int minsize);
-static size_t dhcp_packet_size(struct daemon *daemon, struct dhcp_packet *mess, struct dhcp_netid *netid);
-static void clear_packet(struct dhcp_packet *mess, unsigned char *end);
+static size_t dhcp_packet_size(struct dhcp_packet *mess, struct dhcp_netid *netid);
+static void clear_packet(struct dhcp_packet *mess, unsigned char *end, unsigned char *agent_id);
+static void restore_agent_id(unsigned char *agent_id, struct dhcp_packet *mess, unsigned char *real_end);
 static void do_options(struct dhcp_context *context,
 		       struct dhcp_packet *mess,
 		       unsigned char *real_end, 
 		       unsigned char *req_options,
-		       struct daemon *daemon,
 		       char *hostname,
 		       struct dhcp_netid *netid,
 		       struct in_addr subnet_addr,
@@ -96,8 +99,8 @@ static unsigned char *extended_hwaddr(int hwtype, int hwlen, unsigned char *hwad
 static void match_vendor_opts(unsigned char *opt, struct dhcp_opt *dopt); 
 
 	  
-size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *iface_name, 
-		  size_t sz, time_t now, int unicast_dest)
+size_t dhcp_reply(struct dhcp_context *context, char *iface_name, int int_index,
+		  size_t sz, time_t now, int unicast_dest, int *is_inform)
 {
   unsigned char *opt, *clid = NULL;
   struct dhcp_lease *ltmp, *lease = NULL;
@@ -105,7 +108,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
   struct dhcp_mac *mac;
   struct dhcp_netid_list *id_list;
   int clid_len = 0, ignore = 0, do_classes = 0, selecting = 0;
-  struct dhcp_packet *mess = daemon->dhcp_packet.iov_base;
+  struct dhcp_packet *mess = (struct dhcp_packet *)daemon->dhcp_packet.iov_base;
   unsigned char *end = (unsigned char *)(mess + 1); 
   char *hostname = NULL, *offer_hostname = NULL, *client_hostname = NULL;
   int hostname_auth = 0, borken_opt = 0;
@@ -121,6 +124,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
   unsigned char *agent_id = NULL;
   unsigned char *emac = NULL;
   int emac_len;
+  struct dhcp_netid known_id;
 
   subnet_addr.s_addr = 0;
   
@@ -153,7 +157,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	  
 	  if (expand_buf(&daemon->dhcp_packet, size))
 	    {
-	      mess = daemon->dhcp_packet.iov_base;
+	      mess = (struct dhcp_packet *)daemon->dhcp_packet.iov_base;
 	      end = ((unsigned char *)mess) + size;
 	    }
 	}
@@ -324,6 +328,14 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
   
   config = find_config(daemon->dhcp_conf, context, clid, clid_len, 
 		       mess->chaddr, mess->hlen, mess->htype, NULL);
+
+  /* set "known" tag for known hosts */
+  if (config)
+    {
+      known_id.net = "known";
+      known_id.next = netid;
+      netid = &known_id;
+    }
   
   if (mess_type == 0)
     {
@@ -386,7 +398,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	  else
 	    {
 	      if (!(lease = lease_find_by_client(mess->chaddr, mess->hlen, mess->htype, NULL, 0)) ||
-		  !address_available(context, lease->addr))
+		  !address_available(context, lease->addr, netid))
 		{
 		   if (lease)
 		     {
@@ -394,7 +406,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 		       lease_prune(lease, now);
 		       lease = NULL;
 		     }
-		   if (!address_allocate(context, daemon, &mess->yiaddr, mess->chaddr, mess->hlen, netid, now))
+		   if (!address_allocate(context, &mess->yiaddr, mess->chaddr, mess->hlen, netid, now))
 		     message = _("no address available");
 		}
 	      else
@@ -404,12 +416,9 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	  if (!message && 
 	      !lease && 
 	      (!(lease = lease_allocate(mess->yiaddr))))
-	    {
-	      my_syslog(LOG_WARNING, _("Limit of %d leases exceeded."), daemon->dhcp_max);
-	      message = _("no leases left");
-	    }
-	  
-	  if (!message && !(context = narrow_context(context, mess->yiaddr)))
+	    message = _("no leases left");
+	    
+	  if (!message && !(context = narrow_context(context, mess->yiaddr, netid)))
 	    message = _("wrong network");
 	      
 	  if (!message)
@@ -426,16 +435,17 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	      if (hostname)
 		lease_set_hostname(lease, hostname, daemon->domain_suffix, 1); 
 	      lease_set_expires(lease, 0xffffffff, now); /* infinite lease */
+	      lease_set_interface(lease, int_index);
 	      
-	      clear_packet(mess, end);
-	      do_options(context, mess, end, NULL, daemon, 
+	      clear_packet(mess, end, NULL);
+	      do_options(context, mess, end, NULL,  
 			 hostname, netid, subnet_addr, 0, 0, NULL);
 	    }
 	}
       
-      log_packet(daemon, NULL, logaddr, mess->chaddr, mess->hlen, iface_name, message);
+      log_packet(NULL, logaddr, mess->chaddr, mess->hlen, iface_name, message);
       
-      return message ? 0 : dhcp_packet_size(daemon, mess, netid);
+      return message ? 0 : dhcp_packet_size(mess, netid);
     }
       
   if ((opt = option_find(mess, sz, OPTION_CLIENT_FQDN, 4)))
@@ -504,16 +514,33 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
       if (fqdn_flags != 0 || !client_hostname || hostname_isequal(hostname, client_hostname))
         offer_hostname = hostname;
     }
-  else if (client_hostname && (hostname = strip_hostname(daemon, client_hostname)) && !config)
+  else if (client_hostname)
     {
-      /* Search again now we have a hostname. 
-	 Only accept configs without CLID and HWADDR here, (they won't match)
-	 to avoid impersonation by name. */
-      struct dhcp_config *new = find_config(daemon->dhcp_conf, context, NULL, 0,
-					    mess->chaddr, mess->hlen, 
-					    mess->htype, hostname);
-      if (!have_config(new, CONFIG_CLID) && !have_config(new, CONFIG_HWADDR))
-	config = new;
+      char *d = strip_hostname(client_hostname);
+      if (d)
+	my_syslog(LOG_WARNING, _("Ignoring domain %s for DHCP host name %s"), d, client_hostname);
+      
+      if (strlen(client_hostname) != 0)
+	{
+	  hostname = client_hostname;
+	  if (!config)
+	    {
+	      /* Search again now we have a hostname. 
+		 Only accept configs without CLID and HWADDR here, (they won't match)
+		 to avoid impersonation by name. */
+	      struct dhcp_config *new = find_config(daemon->dhcp_conf, context, NULL, 0,
+						    mess->chaddr, mess->hlen, 
+						    mess->htype, hostname);
+	      if (new && !have_config(new, CONFIG_CLID) && !have_config(new, CONFIG_HWADDR))
+		{
+		  config = new;
+		  /* set "known" tag for known hosts */
+		  known_id.net = "known";
+		  known_id.next = netid;
+		  netid = &known_id;
+		}
+	    }
+	}
     }
   
   if (have_config(config, CONFIG_NETID))
@@ -554,6 +581,15 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	mopt = OPTION_VENDOR_ID;
       else if (vendor->match_type == MATCH_USER)
 	mopt = OPTION_USER_CLASS; 
+      else if (vendor->match_type == MATCH_OPTION)
+	{
+	  if (option_find(mess, sz, vendor->option, 1))
+	    {
+	      vendor->netid.next = netid;
+	      netid = &vendor->netid;
+	    }
+	  continue;
+	}
       else
 	continue;
 
@@ -610,7 +646,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
       if (!(opt = option_find(mess, sz, OPTION_REQUESTED_IP, INADDRSZ)))
 	return 0;
       
-      log_packet(daemon, "DECLINE", option_ptr(opt), emac, emac_len, iface_name, daemon->dhcp_buff);
+      log_packet("DECLINE", option_ptr(opt), emac, emac_len, iface_name, daemon->dhcp_buff);
       
       if (lease && lease->addr.s_addr == option_addr(opt).s_addr)
 	lease_prune(lease, now);
@@ -632,7 +668,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
       return 0;
 
     case DHCPRELEASE:
-      if (!(context = narrow_context(context, mess->ciaddr)) ||
+      if (!(context = narrow_context(context, mess->ciaddr, netid)) ||
 	  !(opt = option_find(mess, sz, OPTION_SERVER_IDENTIFIER, INADDRSZ)) ||
 	  (context->local.s_addr != option_addr(opt).s_addr))
 	return 0;
@@ -642,7 +678,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
       else
 	message = _("unknown lease");
 
-      log_packet(daemon, "RELEASE", &mess->ciaddr, emac, emac_len, iface_name, message);
+      log_packet("RELEASE", &mess->ciaddr, emac, emac_len, iface_name, message);
 	
       return 0;
       
@@ -670,7 +706,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 		  unsigned char *mac = extended_hwaddr(ltmp->hwaddr_type, ltmp->hwaddr_len,
 						       ltmp->hwaddr, ltmp->clid_len, ltmp->clid, &len);
 		  my_syslog(LOG_WARNING, _("not using configured address %s because it is leased to %s"),
-			    addrs, print_mac(daemon, mac, len));
+			    addrs, print_mac(daemon->namebuff, mac, len));
 		}
 	      else
 		{
@@ -690,21 +726,23 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	  
 	  if (conf.s_addr)
 	    mess->yiaddr = conf;
-	  else if (lease && address_available(context, lease->addr))
+	  else if (lease && address_available(context, lease->addr, netid))
 	    mess->yiaddr = lease->addr;
-	  else if (opt && address_available(context, addr) && !lease_find_by_addr(addr) && 
+	  else if (opt && address_available(context, addr, netid) && !lease_find_by_addr(addr) && 
 		   !config_find_by_address(daemon->dhcp_conf, addr))
 	    mess->yiaddr = addr;
-	  else if (!address_allocate(context, daemon, &mess->yiaddr, mess->chaddr, mess->hlen, netid, now))
+	  else if (emac_len == 0)
+	    message = _("no unique-id");
+	  else if (!address_allocate(context, &mess->yiaddr, emac, emac_len, netid, now))
 	    message = _("no address available");      
 	}
       
-      log_packet(daemon, "DISCOVER", opt ? option_ptr(opt) : NULL, emac, emac_len, iface_name, message); 
+      log_packet("DISCOVER", opt ? option_ptr(opt) : NULL, emac, emac_len, iface_name, message); 
 
-      if (message || !(context = narrow_context(context, mess->yiaddr)))
+      if (message || !(context = narrow_context(context, mess->yiaddr, netid)))
 	return 0;
 
-      log_packet(daemon, "OFFER" , &mess->yiaddr, emac, emac_len, iface_name, NULL);
+      log_packet("OFFER" , &mess->yiaddr, emac, emac_len, iface_name, NULL);
 
       if (context->netid.net)
 	{
@@ -712,8 +750,8 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	  netid = &context->netid;
 	}
        
-      time = calc_time(context, config, lease, option_find(mess, sz, OPTION_LEASE_TIME, 4), now);
-      clear_packet(mess, end);
+      time = calc_time(context, config, option_find(mess, sz, OPTION_LEASE_TIME, 4));
+      clear_packet(mess, end, agent_id);
       option_put(mess, end, OPTION_MESSAGE_TYPE, 1, DHCPOFFER);
       option_put(mess, end, OPTION_SERVER_IDENTIFIER, INADDRSZ, ntohl(context->local.s_addr));
       option_put(mess, end, OPTION_LEASE_TIME, 4, time);
@@ -723,10 +761,10 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	  option_put(mess, end, OPTION_T1, 4, (time/2));
 	  option_put(mess, end, OPTION_T2, 4, (time*7)/8);
 	}
-      do_options(context, mess, end, req_options, daemon, offer_hostname, 
+      do_options(context, mess, end, req_options, offer_hostname, 
 		 netid, subnet_addr, fqdn_flags, borken_opt, agent_id);
       
-      return dhcp_packet_size(daemon, mess, netid);
+      return dhcp_packet_size(mess, netid);
       
     case DHCPREQUEST:
       if (ignore || have_config(config, CONFIG_DISABLE))
@@ -765,7 +803,12 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 		return 0;
 	      
 	      if (lease && lease->addr.s_addr != mess->yiaddr.s_addr)
-		message = _("wrong address");
+		{
+		  message = _("wrong address");
+		  /* avoid loops when client brain-dead */
+		  lease_prune(lease, now);
+		  lease = NULL;
+		}
 	    }
 	}
       else
@@ -787,7 +830,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	  mess->yiaddr = mess->ciaddr;
 	}
       
-      log_packet(daemon, "REQUEST", &mess->yiaddr, emac, emac_len, iface_name, NULL);
+      log_packet("REQUEST", &mess->yiaddr, emac, emac_len, iface_name, NULL);
  
       if (!message)
 	{
@@ -799,7 +842,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	      if (context->router.s_addr == config->addr.s_addr)
 		break;
 	  
-	  if (!(context = narrow_context(context, mess->yiaddr)))
+	  if (!(context = narrow_context(context, mess->yiaddr, netid)))
 	    {
 	      /* If a machine moves networks whilst it has a lease, we catch that here. */
 	      message = _("wrong network");
@@ -808,7 +851,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	    }
 	  
 	  /* Check for renewal of a lease which is outside the allowed range. */
-	  else if (!address_available(context, mess->yiaddr) &&
+	  else if (!address_available(context, mess->yiaddr, netid) &&
 		   (!have_config(config, CONFIG_ADDR) || config->addr.s_addr != mess->yiaddr.s_addr))
 	    message = _("address not available");
 	  
@@ -830,7 +873,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	  else if ((ltmp = lease_find_by_addr(mess->yiaddr)) && ltmp != lease)
 	    message = _("address in use");
 	  
-	  else if (!clid && mess->hlen == 0)
+	  else if (emac_len == 0)
 	    message = _("no unique-id");
 	  
 	  else if (!lease)
@@ -844,14 +887,16 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 
       if (message)
 	{
-	  log_packet(daemon, "NAK", &mess->yiaddr, emac, emac_len, iface_name, message);
+	  log_packet("NAK", &mess->yiaddr, emac, emac_len, iface_name, message);
 	  
 	  mess->yiaddr.s_addr = 0;
-	  clear_packet(mess, end);
+	  clear_packet(mess, end, agent_id);
 	  option_put(mess, end, OPTION_MESSAGE_TYPE, 1, DHCPNAK);
 	  option_put(mess, end, OPTION_SERVER_IDENTIFIER, INADDRSZ, 
 		     ntohl(context ? context->local.s_addr : fallback.s_addr));
 	  option_put_string(mess, end, OPTION_MESSAGE, message, borken_opt);
+	  /* DHCPNAK gets agent-id too */
+	  restore_agent_id(agent_id, mess, end);
 	  /* This fixes a problem with the DHCP spec, broadcasting a NAK to a host on 
 	     a distant subnet which unicast a REQ to us won't work. */
 	  if (!unicast_dest || mess->giaddr.s_addr != 0 || 
@@ -874,9 +919,8 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 		   /* If the user-class option started as counted strings, the first byte will be zero. */
 		   if (len != 0 && ucp[0] == 0)
 		     ucp++, len--;
-		   if (lease->userclass)
-		     free(lease->userclass);
-		   if ((lease->userclass = malloc(len+1)))
+		   free(lease->userclass);
+		   if ((lease->userclass = whine_malloc(len+1)))
 		     {
 		       memcpy(lease->userclass, ucp, len);
 		       lease->userclass[len] = 0;
@@ -887,9 +931,8 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 		 {
 		   int len = option_len(opt);
 		   unsigned char *ucp = option_ptr(opt);
-		   if (lease->vendorclass)
-		     free(lease->vendorclass);
-		   if ((lease->vendorclass = malloc(len+1)))
+		   free(lease->vendorclass);
+		   if ((lease->vendorclass = whine_malloc(len+1)))
 		     {
 		       memcpy(lease->vendorclass, ucp, len);
 		       lease->vendorclass[len] = 0;
@@ -898,7 +941,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 		 }
 	     }
 	   
-	   if (!hostname_auth && (client_hostname = host_from_dns(daemon, mess->yiaddr)))
+	   if (!hostname_auth && (client_hostname = host_from_dns(mess->yiaddr)))
 	     {
 	      hostname = client_hostname;
 	      hostname_auth = 1;
@@ -910,7 +953,7 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	      netid = &context->netid;
 	    }
 	
-	  time = calc_time(context, config, NULL, option_find(mess, sz, OPTION_LEASE_TIME, 4), now);
+	  time = calc_time(context, config, option_find(mess, sz, OPTION_LEASE_TIME, 4));
 	  lease_set_hwaddr(lease, mess->chaddr, clid, mess->hlen, mess->htype, clid_len);
 	   
 	  /* if all the netids in the ignore_name list are present, ignore client-supplied name */
@@ -926,10 +969,11 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	    lease_set_hostname(lease, hostname, daemon->domain_suffix, hostname_auth);
 	  
 	  lease_set_expires(lease, time, now);
-	  	
-	  log_packet(daemon, "ACK", &mess->yiaddr, emac, emac_len, iface_name, hostname);  
+	  lease_set_interface(lease, int_index);
+	
+	  log_packet("ACK", &mess->yiaddr, emac, emac_len, iface_name, hostname);  
 	  
-	  clear_packet(mess, end);
+	  clear_packet(mess, end, agent_id);
 	  option_put(mess, end, OPTION_MESSAGE_TYPE, 1, DHCPACK);
 	  option_put(mess, end, OPTION_SERVER_IDENTIFIER, INADDRSZ, ntohl(context->local.s_addr));
 	  option_put(mess, end, OPTION_LEASE_TIME, 4, time);
@@ -940,39 +984,59 @@ size_t dhcp_reply(struct daemon *daemon, struct dhcp_context *context, char *ifa
 	      option_put(mess, end, OPTION_T1, 4, (time/2) - fuzz);
 	      option_put(mess, end, OPTION_T2, 4, ((time/8)*7) - fuzz);
 	    }
-	  do_options(context, mess, end, req_options, daemon, hostname, 
+	  do_options(context, mess, end, req_options, hostname, 
 		     netid, subnet_addr, fqdn_flags, borken_opt, agent_id);
 	}
 
-      return dhcp_packet_size(daemon, mess, netid); 
+      return dhcp_packet_size(mess, netid); 
       
     case DHCPINFORM:
       if (ignore || have_config(config, CONFIG_DISABLE))
 	message = _("ignored");
       
-      log_packet(daemon, "INFORM", &mess->ciaddr, emac, emac_len, iface_name, message);
+      log_packet("INFORM", &mess->ciaddr, emac, emac_len, iface_name, message);
      
       if (message || mess->ciaddr.s_addr == 0 || 
-	  !(context = narrow_context(context, mess->ciaddr)))
+	  !(context = narrow_context(context, mess->ciaddr, netid)))
 	return 0;
       
-      log_packet(daemon, "ACK", &mess->ciaddr, emac, emac_len, iface_name, hostname);
+      /* Find a least based on IP address if we didn't
+	 get one from MAC address/client-d */
+      if (!lease &&
+	  (lease = lease_find_by_addr(mess->ciaddr)) && 
+	  lease->hostname)
+	hostname = lease->hostname;
+      
+      if (!hostname)
+	hostname = host_from_dns(mess->ciaddr);
 
+      log_packet("ACK", &mess->ciaddr, emac, emac_len, iface_name, hostname);
+      
       if (context->netid.net)
 	{
 	  context->netid.next = netid;
 	  netid = &context->netid;
 	}
        
-      clear_packet(mess, end);
+      clear_packet(mess, end, agent_id);
       option_put(mess, end, OPTION_MESSAGE_TYPE, 1, DHCPACK);
       option_put(mess, end, OPTION_SERVER_IDENTIFIER, INADDRSZ, ntohl(context->local.s_addr));
-      if (!hostname)
-	hostname = host_from_dns(daemon, mess->yiaddr);
-      do_options(context, mess, end, req_options, daemon, hostname, 
+
+      if (lease)
+	{
+	  if (lease->expires == 0)
+	    time = 0xffffffff;
+	  else
+	    time = (unsigned int)difftime(lease->expires, now);
+	  option_put(mess, end, OPTION_LEASE_TIME, 4, time);
+	  lease_set_interface(lease, int_index);
+	}
+
+      do_options(context, mess, end, req_options, hostname, 
 		 netid, subnet_addr, fqdn_flags, borken_opt, agent_id);
       
-      return dhcp_packet_size(daemon, mess, netid); 
+      *is_inform = 1; /* handle reply differently */
+      return dhcp_packet_size(mess, netid); 
     }
   
   return 0;
@@ -1013,8 +1077,7 @@ static unsigned char *extended_hwaddr(int hwtype, int hwlen, unsigned char *hwad
   return hwaddr;
 }
 
-static unsigned int calc_time(struct dhcp_context *context, struct dhcp_config *config, 
-			      struct dhcp_lease *lease, unsigned char *opt, time_t now)
+static unsigned int calc_time(struct dhcp_context *context, struct dhcp_config *config, unsigned char *opt)
 {
   unsigned int time = have_config(config, CONFIG_TIME) ? config->lease_time : context->lease_time;
   
@@ -1026,17 +1089,6 @@ static unsigned int calc_time(struct dhcp_context *context, struct dhcp_config *
       if (time == 0xffffffff || (req_time != 0xffffffff && req_time < time))
 	time = req_time;
     }
-  else if (lease && lease->expires != 0 && difftime(lease->expires, now) > 0.0)
-    {
-      unsigned int lease_time = (unsigned int)difftime(lease->expires, now);
-      
-      /* put a floor on lease-remaining time. */
-      if (lease_time < 360 )
-	lease_time = 360;
-      
-      if (time > lease_time)
-	time = lease_time;
-    }
 
   return time;
 }
@@ -1056,7 +1108,7 @@ static int sanitise(unsigned char *opt, char *buf)
   for (i = option_len(opt); i > 0; i--)
     {
       char c = *p++;
-      if (isprint(c))
+      if (isprint((int)c))
 	*buf++ = c;
     }
   *buf = 0; /* add terminator */
@@ -1064,8 +1116,8 @@ static int sanitise(unsigned char *opt, char *buf)
   return 1;
 }
 
-static void log_packet(struct daemon *daemon, char *type, void *addr, 
-		       unsigned char *ext_mac, int mac_len, char *interface, char *string)
+static void log_packet(char *type, void *addr, unsigned char *ext_mac, 
+		       int mac_len, char *interface, char *string)
 {
   struct in_addr a;
 
@@ -1079,11 +1131,11 @@ static void log_packet(struct daemon *daemon, char *type, void *addr,
 	    interface, 
 	    addr ? inet_ntoa(a) : "",
 	    addr ? " " : "",
-	    print_mac(daemon, ext_mac, mac_len),
+	    print_mac(daemon->namebuff, ext_mac, mac_len),
 	    string ? string : "");
 }
 
-static void log_options(struct daemon *daemon, unsigned char *start)
+static void log_options(unsigned char *start)
 {
   while (*start != OPTION_END)
     {
@@ -1093,7 +1145,7 @@ static void log_options(struct daemon *daemon, unsigned char *start)
 		start[1], start[0],
 		text ? ":" : "", text ? text : "",
 		start[1] == 0 ? "" : "  ",
-		start[1] == 0 ? "" : print_mac(daemon, &start[2], trunc),
+		start[1] == 0 ? "" : print_mac(daemon->namebuff, &start[2], trunc),
 		trunc == start[1] ? "" : "...");
       start += start[1] + 2;
     }
@@ -1194,11 +1246,12 @@ static unsigned char *find_overload(struct dhcp_packet *mess)
   return NULL;
 }
 
-static size_t dhcp_packet_size(struct daemon *daemon, struct dhcp_packet *mess, struct dhcp_netid *netid)
+static size_t dhcp_packet_size(struct dhcp_packet *mess, struct dhcp_netid *netid)
 {
   unsigned char *p = dhcp_skip_opts(&mess->options[0] + sizeof(u32));
   unsigned char *overload;
   size_t ret;
+  struct dhcp_netid_list *id_list;
 
   /* We do logging too */
   if (netid && (daemon->options & OPT_LOG_OPTS))
@@ -1222,19 +1275,24 @@ static size_t dhcp_packet_size(struct daemon *daemon, struct dhcp_packet *mess,
 	{
 	  *dhcp_skip_opts(mess->file) = OPTION_END;
 	  if (daemon->options & OPT_LOG_OPTS)
-	    log_options(daemon, mess->file);
+	    log_options(mess->file);
 	}
       if (option_uint(overload, 1) & 2)
 	{
 	  *dhcp_skip_opts(mess->sname) = OPTION_END;
 	  if (daemon->options & OPT_LOG_OPTS)
-	    log_options(daemon, mess->sname);
+	    log_options(mess->sname);
 	}
     }
 
   *p++ = OPTION_END;
+  
   if (daemon->options & OPT_LOG_OPTS)
-    log_options(daemon, &mess->options[0] + sizeof(u32));
+    log_options(&mess->options[0] + sizeof(u32));
+  
+  for (id_list = daemon->force_broadcast; id_list; id_list = id_list->next)
+    if (match_netid(id_list->list, netid, 0))
+      mess->flags |= htons(0x8000); /* force broadcast */
   
   ret = (size_t)(p - (unsigned char *)mess);
   
@@ -1407,19 +1465,33 @@ static void match_vendor_opts(unsigned char *opt, struct dhcp_opt *dopt)
     }
 }
 
-static void clear_packet(struct dhcp_packet *mess, unsigned char *end)
+static void clear_packet(struct dhcp_packet *mess, unsigned char *end, unsigned char *agent_id)
 {
+  /* don't clear agent_id */
+  if (agent_id)
+    end = agent_id;
+
   memset(mess->sname, 0, sizeof(mess->sname));
   memset(mess->file, 0, sizeof(mess->file));
   memset(&mess->options[0] + sizeof(u32), 0, end - (&mess->options[0] + sizeof(u32)));
   mess->siaddr.s_addr = 0;
 }
 
+static void restore_agent_id(unsigned char *agent_id, struct dhcp_packet *mess, unsigned char *real_end)
+{ 
+  if (agent_id)
+    {
+      unsigned char *p = dhcp_skip_opts(&mess->options[0] + sizeof(u32));
+      memmove(p, agent_id, real_end - agent_id);
+      p += real_end - agent_id;
+      memset(p, 0, real_end - p); /* in case of overlap */
+    }
+}
+
 static void do_options(struct dhcp_context *context,
 		       struct dhcp_packet *mess,
 		       unsigned char *real_end, 
 		       unsigned char *req_options,
-		       struct daemon *daemon,
 		       char *hostname,
 		       struct dhcp_netid *netid,
 		       struct in_addr subnet_addr,
@@ -1432,6 +1504,7 @@ static void do_options(struct dhcp_context *context,
   unsigned char *p, *end = agent_id ? agent_id : real_end;
   int i, len, force_encap = 0;
   unsigned char f0 = 0, s0 = 0;
+  int done_file = 0, done_server = 0;
 
   /* logging */
   if ((daemon->options & OPT_LOG_OPTS) && req_options)
@@ -1440,12 +1513,12 @@ static void do_options(struct dhcp_context *context,
       for (i = 0; req_options[i] != OPTION_END; i++)
 	{
 	  char *s = option_string(req_options[i]);
-	  q +=snprintf(q, MAXDNAME - (q - daemon->namebuff),
-		       "%d%s%s%s", 
-		       req_options[i],
-		       s ? ":" : "",
-		       s ? s : "", 
-		       req_options[i+1] == OPTION_END ? "" : ", ");
+	  q += snprintf(q, MAXDNAME - (q - daemon->namebuff),
+			"%d%s%s%s", 
+			req_options[i],
+			s ? ":" : "",
+			s ? s : "", 
+			req_options[i+1] == OPTION_END ? "" : ", ");
 	  if (req_options[i+1] == OPTION_END || (q - daemon->namebuff) > 40)
 	    {
 	      q = daemon->namebuff;
@@ -1468,33 +1541,30 @@ static void do_options(struct dhcp_context *context,
   
   /* See if we can send the boot stuff as options.
      To do this we need a requested option list, BOOTP
-     and very old DHCP clients won't have this. 
+     and very old DHCP clients won't have this, we also 
+     provide an manual option to disable it.
      Some PXE ROMs have bugs (surprise!) and need zero-terminated 
      names, so we always send those. */
   if (boot)
     {
       if (boot->sname)
-	{
-	  if (req_options && in_list(req_options, OPTION_SNAME))
+	{	  
+	  if (!(daemon->options & OPT_NO_OVERRIDE) &&
+	      req_options && 
+	      in_list(req_options, OPTION_SNAME))
 	    option_put_string(mess, end, OPTION_SNAME, boot->sname, 1);
 	  else
-	    {
-	      if (daemon->options & OPT_LOG_OPTS)
-		my_syslog(LOG_INFO, _("server name: %s"), boot->sname);
-	      strncpy((char *)mess->sname, boot->sname, sizeof(mess->sname)-1);
-	    }
+	    strncpy((char *)mess->sname, boot->sname, sizeof(mess->sname)-1);
 	}
       
       if (boot->file)
 	{
-	  if (req_options && in_list(req_options, OPTION_FILENAME))
+	  if (!(daemon->options & OPT_NO_OVERRIDE) &&
+	      req_options && 
+	      in_list(req_options, OPTION_FILENAME))
 	    option_put_string(mess, end, OPTION_FILENAME, boot->file, 1);
 	  else
-	    {
-	      if (daemon->options & OPT_LOG_OPTS)
-		my_syslog(LOG_INFO, _("bootfile name: %s"), boot->file);
-	      strncpy((char *)mess->file, boot->file, sizeof(mess->file)-1);
-	    }
+	    strncpy((char *)mess->file, boot->file, sizeof(mess->file)-1);
 	}
       
       if (boot->next_server.s_addr)
@@ -1503,12 +1573,39 @@ static void do_options(struct dhcp_context *context,
       if (daemon->options & OPT_LOG_OPTS)
 	my_syslog(LOG_INFO, _("next server: %s"), inet_ntoa(mess->siaddr));
     }
+  else
+    /* Use the values of the relevant options if no dhcp-boot given and
+       they're no explicitly asked for as options. */
+    {
+      if ((!req_options || !in_list(req_options, OPTION_FILENAME)) &&
+	  (opt = option_find2(netid, config_opts, OPTION_FILENAME)))
+	{
+	  strncpy((char *)mess->file, (char *)opt->val, sizeof(mess->file)-1);
+	  done_file = 1;
+	}
+
+      if ((!req_options || !in_list(req_options, OPTION_SNAME)) &&
+	  (opt = option_find2(netid, config_opts, OPTION_SNAME)))
+	{
+	  strncpy((char *)mess->sname, (char *)opt->val, sizeof(mess->sname)-1);
+	  done_server = 1;
+	}
+    }
   
+  if (daemon->options & OPT_LOG_OPTS)
+    { 
+      if (strlen((char *)mess->file) != 0)
+	my_syslog(LOG_INFO, _("bootfile name: %s"), (char *)mess->file);
+      
+      if (strlen((char *)mess->sname) != 0)
+	my_syslog(LOG_INFO, _("server name: %s"), (char *)mess->sname);
+    } 
+      
   /* We don't want to do option-overload for BOOTP, so make the file and sname
      fields look like they are in use, even when they aren't. This gets restored
      at the end of this function. */
 
-  if (!req_options)
+  if (!req_options || (daemon->options & OPT_NO_OVERRIDE))
     {
       f0 = mess->file[0];
       mess->file[0] = 1;
@@ -1549,12 +1646,11 @@ static void do_options(struct dhcp_context *context,
       !option_find2(netid, config_opts, OPTION_DOMAINNAME))
     option_put_string(mess, end, OPTION_DOMAINNAME, daemon->domain_suffix, null_term);
  
-  /* Note that we ignore attempts to set the hostname using 
-     --dhcp-option=12,<name> and the fqdn using
-     --dhc-option=81,<name> */
+  /* Note that we ignore attempts to set the fqdn using --dhc-option=81,<name> */
   if (hostname)
     {
-      if (in_list(req_options, OPTION_HOSTNAME))
+      if (in_list(req_options, OPTION_HOSTNAME) &&
+	  !option_find2(netid, config_opts, OPTION_HOSTNAME))
 	option_put_string(mess, end, OPTION_HOSTNAME, hostname, null_term);
       
       if (fqdn_flags != 0)
@@ -1600,41 +1696,54 @@ static void do_options(struct dhcp_context *context,
 
   for (opt = config_opts; opt; opt = opt->next)
     {
+      int optno = opt->opt;
+
       /* was it asked for, or are we sending it anyway? */
-      if (!(opt->flags & DHOPT_FORCE) && !in_list(req_options, opt->opt))
+      if (!(opt->flags & DHOPT_FORCE) && !in_list(req_options, optno))
 	continue;
       
       /* prohibit some used-internally options */
-      if (opt->opt == OPTION_HOSTNAME ||
-	  opt->opt == OPTION_CLIENT_FQDN ||
-	  opt->opt == OPTION_MAXMESSAGE ||
-	  opt->opt == OPTION_OVERLOAD ||
-	  opt->opt == OPTION_PAD ||
-	  opt->opt == OPTION_END)
+      if (optno == OPTION_CLIENT_FQDN ||
+	  optno == OPTION_MAXMESSAGE ||
+	  optno == OPTION_OVERLOAD ||
+	  optno == OPTION_PAD ||
+	  optno == OPTION_END)
+	continue;
+
+      if (optno == OPTION_SNAME && done_server)
+	continue;
+
+      if (optno == OPTION_FILENAME && done_file)
 	continue;
       
       /* netids match and not encapsulated? */
-      if (opt != option_find2(netid, config_opts, opt->opt))
+      if (opt != option_find2(netid, config_opts, optno))
 	continue;
       
       /* For the options we have default values on
 	 dhc-option=<optionno> means "don't include this option"
 	 not "include a zero-length option" */
       if (opt->len == 0 && 
-	  (opt->opt == OPTION_NETMASK ||
-	   opt->opt == OPTION_BROADCAST ||
-	   opt->opt == OPTION_ROUTER ||
-	   opt->opt == OPTION_DNSSERVER))
+	  (optno == OPTION_NETMASK ||
+	   optno == OPTION_BROADCAST ||
+	   optno == OPTION_ROUTER ||
+	   optno == OPTION_DNSSERVER || 
+	   optno == OPTION_DOMAINNAME ||
+	   optno == OPTION_HOSTNAME))
 	continue;
+      
+      /* always force null-term for filename ans servername - buggy PXE again. */
+      len = do_opt(opt, NULL, context->local, 
+		   (optno == OPTION_SNAME || optno == OPTION_FILENAME) ? 1 : null_term);
 
-      len = do_opt(opt, NULL, context->local, null_term);
-      if ((p = free_space(mess, end, opt->opt, len)))
+      if ((p = free_space(mess, end, optno, len)))
 	{
-	  do_opt(opt, p, context->local, null_term);
-
+	  do_opt(opt, p, context->local, 
+		 (optno == OPTION_SNAME || optno == OPTION_FILENAME) ? 1 : null_term);
+	  
 	  /* If we send a vendor-id, revisit which vendor-ops we consider 
 	     it appropriate to send. */
-	  if (opt->opt == OPTION_VENDOR_ID)
+	  if (optno == OPTION_VENDOR_ID)
 	    match_vendor_opts(p - 2, config_opts);
 	}  
     }
@@ -1693,16 +1802,10 @@ static void do_options(struct dhcp_context *context,
     }
   
   /* move agent_id back down to the end of the packet */
-  if (agent_id)
-    {
-      p = dhcp_skip_opts(&mess->options[0] + sizeof(u32));
-      memmove(p, agent_id, real_end - agent_id);
-      p += real_end - agent_id;
-      memset(p, 0, real_end - p); /* in case of overlap */
-    }
+  restore_agent_id(agent_id, mess, real_end);
 
   /* restore BOOTP anti-overload hack */
-  if (!req_options)
+  if (!req_options || (daemon->options & OPT_NO_OVERRIDE))
     {
       mess->file[0] = f0;
       mess->sname[0] = s0;
